(****************************************************************************

  Name      : ZIP File System
  Version   : 0.3b
  Part Of   : DIABOLO Game Engine
  Compiler  : Free Pascal 0.99.5 (DJGPP v2)
  Date      : 10-6-98
  Author    : Ivn Montes Velencoso <imv@SoftHome.net>
              The ZIP uncompression stuff ripped from <stefan.goehler@gmx.de>
  Purpose   : Handling of files like Quake 3 do, from the Disk or from a ZIP.

  History   : 24-MAY-98 : Support for Quake II PACK files
              06-JUN-98 : Now works with compressed ZIP files (methods 0..1)
              not Huffman method (8) yet.
              09-JUN-98 : Full featured Blockread

  ToDo      : -Implement a good error checking!!!!!
              -Good examples

 A brief explanation :

    With this unit you can access to files from the disk or files packed
    inside ZIP files with a total transparency to the programmer.
    For that this unit uses a set of modificated routines based on the
    pascal standard file handling functions : ASSIGN, CLOSE, BLOCKREAD, SEEK,
    EOF.



ZFSResult Values
----------------

0      : All O.K.
1..255 : The same as IOResult of the Pascal Run Time Library
256    : The pointer is NIL and can not be erased
257    : Unable to create a new node in a linked list
258    : No nodes to be delete in this linked list
259    : Node not found in this linked list
300    : The pack file hasn't got files
301    : No openned pack files
302    :
303    : NewFile is not openned
304    : Error seeking the ZIP file
305    : Not enough memory for the ZIP file buffer
306    : Missing or invalid local file header in Zip file
307    : Filename of compressed file exceeds 255 chars!
308    : Extra field of compressed file exceeds 255 chars!
310    : Invaled code encountered at LZW decompression (SHRINK)

*****************************************************************************)

Unit ZFS;

INTERFACE

(************************ ZIP STRUCTURES ****************************)

type
  Local_File_Header_Type = record
    {Signature              :LongInt; }
    Extract_Version_Reqd   :Word;
    Bit_Flag               :Word;
    Compress_Method        :Word;
    Last_Mod_Time          :Word;
    Last_Mod_Date          :Word;
    Crc32                  :LongInt;
    Compressed_Size        :LongInt;
    Uncompressed_Size      :LongInt;
    Filename_Length        :Word;
    Extra_Field_Length     :Word;
  end;



{ Buffer definition }
const BUFSIZE       = 32768;           { Size of buffers for I/O }
type
  BufPtr        = ^BufType;
  BufType       = array[1..BUFSIZE] of byte;


{ Stuff needed for unSHRINKing }
const
  MINCODESIZE    =   9;
  MAXCODESIZE    =  13;
  SPECIAL        = 256;
  FIRSTFREE      = 257;
  LZW_Table_SIZE = (1 shl MAXCODESIZE)-1;      { 0..8191 }
  LZW_Stack_SIZE = (1 shl MAXCODESIZE)-1;      { 0..8191 }

type
  LZW_Table_Rec  =  record
    Prefix      :integer;
    Suffix      :byte;
    ChildCount  :word;  { If ChildCount = 0 then leaf node }
  end;
  LZW_Table_Ptr  =  ^LZW_Table_Type;
  LZW_Table_Type =  array[0..LZW_Table_SIZE] of LZW_Table_Rec;

  FreeListPtr    =  ^FreeListArray;
  FreeListArray  =  array[FIRSTFREE..LZW_Table_SIZE] of Word;

  StackPtr       =  ^StackType;
  StackType      =  array[0..LZW_Stack_SIZE] of word;



type
  ZipObj =Object
   PRIVATE
    ZipName       :string;         { Name of Zip file to be processed       }
    ZipFile       :file;           { Zip file variable                      }
    EndFile       :boolean;        { End of file indicator for ZipFile      }
    ZipBuf        :BufPtr;         { Input buffer for ZipFile               }
    ZipPtr        :word;           { Index for ZipFile input buffer         }
    ZipCount      :word;           { Count of bytes in ZipFile input buffer }
    Dest          :pointer;
    ExtCount      :longint;        { Count of characters written to output  }
    LocalHdr      :Local_File_Header_Type;  { Storage for a local file hdr  }
    Hdr_FileName  :string;
    Hdr_ExtraField:string;
    Hdr_Comment   :string;

    Bytes_To_Go   :longint;        { Bytes left to process in compressed file}


    { Stuff needed for unSHRINKing }
    LZW_Table   :LZW_Table_Ptr; { Code table for LZW decoding               }
    FreeList    :FreeListPtr;   { List of free table entries                }
    NextFree    :word;          { Index for free list array                 }
                                {   FreeList^[NextFree] always contains the }
                                {   index of the next available entry in    }
                                {   the LZW Prefix:Suffix table (LZW_Table^)}
    LZW_Stack   :StackPtr;      { A stack used to build decoded strings     }
    StackIdx    :word;          { Stack array index variable                }
                                {   StackIdx always points to the next      }
                                {   available entry in the stack            }
    SaveByte    :byte;          { Our input code buffer - 1 byte long       }
    BitsLeft    :byte;          { Unprocessed bits in the input code buffer }
    FirstCh     :boolean;       { Flag indicating first char being processed}

    function    UnZip(count, btg :longint) :longint;
    constructor Init(fname:string);
    destructor  Close;

    function    Read_Local_Hdr : boolean;

    procedure   Read_Zip_Block;

    function    Get_Zip : Integer;
    procedure   Put_Ext(C : Byte);

    procedure   fSkip(Offset : longInt);
    procedure   fRead(var Buf; RecLen :word);

    function    Get_Compressed : integer;
    function    Get_Code(CodeSize : Byte) : Integer;

    function    LZW_Init : boolean;
    procedure   LZW_Cleanup;
    procedure   LZW_Clear_Table;
    procedure   LZW_Add_To_Table(Prefix : integer; Suffix : byte);
    procedure   UnShrink(count:longint);
  end;


TYPE
  tFileZIP = record {Packed Files info}
                Nom   :string;      {file name}
                Kind  :byte;        {0:Store, 1:LZW, 2:..., 8:HUFFMAN}
                Pos   :longint;     {file position in the PACK}
                Size  :longint;     {file size}
                USize :longint;     {Uncompressed size}
  end; {tfilepack}

  pMEMDir = ^tMEMDir;
  tMEMDir = record
              Obj  :tFileZIP;
              next :pMEMDir;
  end; {tMEMDir}


  pZIP = ^tZIP;
  tZIP = record
            Active      :Boolean;      {Is this active}
            Name        :String;       {file name .ZIP}
            NumFiles    :Word;         {number of files zipped}
            FirstMEMDir :pMEMDir;      {Index of the files linked list}
            Obj         :ZipObj;
            Next        :pZIP;         {Next PACK file (NIL if none)}
  end; {tpack}


  ZFSfile = record
              InZIP   :Boolean;       {Is in ZIP?}
              Openned :Boolean;       {Is currently openned}
              handle  :file;          {file handle, in case it is on disk}
              ZIP     :pZIP;          {ZIP where it is}
              comp    :byte;          {Type of compression}
              pos     :longint;       {position in case it is on PACK}
              size    :longint;       {size in case it is on PACK}
              usize   :longint;       {size of the uncompressed archive}
              curr    :longint;       {Current position in the file}

  end;


const
  FirstZIP   :pZIP = NIL;   {Index of the PACK linked list}

  ZFSResult  :Word = 0;


function  FormatFileName(s:string) : string;
{Converts a LINUX filename to a DOS one}

Function  OpenZIP (ZIPName :string ):pZIP;
{Opens a ZIP}
Procedure CloseZIP(zipf:pZip);
{Closes a ZIP}
Procedure KillZIPs;
{Closes all the ZIPs}

function  FileExist( sExp :STRING ): BOOLEAN;
{Returns TRUE if the file exists}
Procedure Assign(var f:ZFSfile; fname:string);
{Returns TRUE if the file is open correctly}
Procedure Close(var f:ZFSfile);
{Close an openned file}
procedure seek(var f:ZFSfile; N:longint);
{Seeks a file}
Function  EOF(f:ZFSfile):Boolean;
{Returns true if the file have reached the end}
Function  BlockRead(var F :ZFSfile; var Buf; Count :longint):longint;
{Read from a file, returns the bytes read}


IMPLEMENTATION

uses Crt, DOS;

var
  OldExitProc :Pointer;     {Stores temporaly ExitProc}


(************************************************************************)
(******************** ZIP HANDLING ROUTINES *****************************)
(************************************************************************)

type
  Central_File_Header_Type = record
    {Signature            :LongInt; }
    MadeBy_Version       :word;
    Extract_Version_Reqd :word;
    Bit_Flag             :word;
    Compress_Method      :word;
    Last_Mod_Time        :word;
    Last_Mod_Date        :word;
    Crc32                :longint;
    Compressed_Size      :longint;
    Uncompressed_Size    :longint;
    Filename_Length      :word;
    Extra_Field_Length   :word;
    File_Comment_Length  :word;
    Starting_Disk_Num    :word;
    Internal_Attributes  :word;
    External_Attributes  :longint;
    Local_Header_Offset  :longint;
  end;

  End_of_Central_Dir_Type =  record
    {Signature               :LongInt; }
    Disk_Number             :word;
    Central_Dir_Start_Disk  :word;
    Entries_This_Disk       :word;
    Total_Entries           :word;
    Central_Dir_Size        :longint;
    Start_Disk_Offset       :longint;
    ZipFile_Comment_Length  :word;
  end;

{ Define ZIP file header types }
const
  LOCAL_FILE_HEADER_SIGNATURE = $04034B50;
  CENTRAL_FILE_HEADER_SIGNATURE = $02014B50;
  END_OF_CENTRAL_DIR_SIGNATURE = $06054B50;



{INITIALIZES A ZIP OBJECT}
Constructor ZipObj.Init(fname:string);
begin
  ZFSResult:=0;
  system.assign( ZipFile, fName);
  system.FileMode := 64;  { Read Only }
  {$I-}system.reset(ZipFile, 1);{$I+}
  if system.IOResult<>0 then
   begin
     ZFSResult:=IOResult;
     exit;
   end;
  ZipName := fname;
  EndFile := false;

  getmem(ZipBuf, SizeOf(ZipBuf^));
  if (ZipBuf = nil) then ZFSResult:=305; {Not enough memory};

  Dest:=NIL;

  Read_Zip_Block;
end;



procedure ZipObj.Read_Zip_Block;
begin
  {$I-} system.blockread(ZipFile, ZipBuf^, BUFSIZE, ZipCount); {$I+}
  ZFSResult:=IOResult;
  If ZipCount = 0 then EndFile := true;
  ZipPtr := 1;
end {Read_Zip_Block};


function ZipObj.Get_Zip : Integer;
begin
  if ZipPtr > ZipCount then Read_Zip_Block;

  if EndFile then
   Get_Zip := -1
  else
   begin
     Get_Zip := ZipBuf^[ZipPtr];
     inc(ZipPtr);
   end {if};
end;

procedure ZipObj.Put_Ext(C : Byte);
begin
  mem[ofs(dest^)+ExtCount]:=c;

  inc(ExtCount);
end;


destructor ZipObj.Close;
begin
  ZFSResult:=0;
  if ZipBuf<>NIL then freemem(ZipBuf, SizeOf(ZipBuf^));

  {$I-}System.Close(ZipFile);{$I+}
  ZFSResult:=IOResult;
end;


procedure ZipObj.fSkip(Offset : longInt);
var
  Rec :longInt;
begin
  if (Offset + ZipPtr) <= ZipCount then
   inc(ZipPtr, Offset)
  else
   begin
     Rec := system.FilePos(ZipFile) + (Offset - (ZipCount - ZipPtr) - 1);
     {$I-}system.Seek(ZipFile, Rec);{$I+}
     if system.IOResult<>0 then
      begin
        ZFSResult:=304;
        exit;
      end;
     Read_Zip_Block;
   end {if};
end;


procedure ZipObj.fRead(var Buf; RecLen :word);
var
  I  :word;
  B  :array[1..MaxInt] of byte absolute Buf;
begin
  for i:=1 to RecLen do b[i] := Get_Zip;
end;


function ZipObj.Read_Local_Hdr : boolean;
var
  Sig : longInt;
begin
  if EndFile then
   Read_Local_Hdr := false
  else
   begin
     fRead(Sig, SizeOf(Sig));
     if Sig = CENTRAL_FILE_HEADER_SIGNATURE then
      begin
        Read_Local_Hdr := false;
      end {then}
     else
      begin
        if Sig <> LOCAL_FILE_HEADER_SIGNATURE then
         begin
           ZFSResult:=306;  {Missing or invalid local file header in Zip file}
           EXIT;
         end;
        fread(LocalHdr, SizeOf(LocalHdr));
        with LocalHdr do
         begin
           if FileName_Length > 255 then
            begin
              ZFSResult:=307; {Filename of compressed file exceeds 255 chars!};
              EXIT;
            end;
           fread(Hdr_FileName[1], FileName_Length);
           Hdr_FileName[0] := Chr(FileName_Length);
           if Extra_Field_Length > 255 then
            begin
              ZFSResult:=308;
              EXIT;
            end;
           fread(Hdr_ExtraField[1], Extra_Field_Length);
           Hdr_ExtraField[0] := Chr(Extra_Field_Length);
         end {with};
        Read_Local_Hdr := true;
      end {if};
   end {if};
end;


function ZipObj.Get_Compressed : integer;
begin
  if Bytes_To_Go < 1 then
   Get_Compressed := -1 {end of the descompression}
  else
   begin
     Get_Compressed := Get_Zip; {get another byte from the ZIP}
     Dec(Bytes_To_Go);
   end {if};
end;


function ZipObj.LZW_Init : boolean;
var
  I : word;
begin
  new(LZW_Table);
  If LZW_Table = nil then
   begin
     LZW_Init := false;
     exit;
   end;
  for I := 0 to LZW_Table_SIZE do
   begin
     with LZW_Table^[I] do
      begin
        Prefix     := -1;
        if I < 256 then Suffix := i else Suffix  := 0;
        ChildCount := 0;
      end {with};
   end {for};

  getmem(FreeList, SizeOf(FreeList^));
  if FreeList = nil then
   begin
     LZW_Init := FALSE;
     exit;
   end;
  for I := FIRSTFREE to LZW_Table_SIZE do FreeList^[I] := I;
  NextFree := FIRSTFREE;

  { Initialize the LZW Character Stack }
  new(LZW_Stack);
  if LZW_Stack = nil then
   begin
     LZW_Init := false;
     exit;
   end;
  StackIdx := 0;
  LZW_Init := true;
end;


procedure ZipObj.LZW_Cleanup;
begin
  freemem(LZW_Table,sizeof(LZW_Table^));
  freemem(FreeList,sizeof(FreeList^));
  freemem(LZW_Stack,sizeof(LZW_Stack^));
end;


procedure ZipObj.LZW_Clear_Table;
var
  i :  word;
begin
  StackIdx := 0;
  for i := FIRSTFREE to LZW_Table_SIZE do
   begin      { Find all leaf nodes }
     if LZW_Table^[I].ChildCount = 0 then
      begin
        LZW_Stack^[StackIdx] := i;              { and put each on stack }
        inc(StackIdx);
      end {if};
   end {for};

  NextFree := succ(LZW_Table_SIZE);

  while StackIdx > 0 do
   begin                        { clear all leaf nodes }
     dec(StackIdx);
     i := LZW_Stack^[StackIdx];
     with LZW_Table^[I] do
      begin
        if LZW_Table^[I].Prefix <> -1 then dec(LZW_Table^[Prefix].ChildCount);
        Prefix     := -1;
        Suffix     :=  0;
        ChildCount :=  0;
      end {with};
     Dec(NextFree);                { add cleared nodes to FreeList }
     FreeList^[NextFree] := I;
   end {while};
end;


procedure ZipObj.LZW_Add_To_Table(Prefix : integer; Suffix : byte);
var
  i : word;
begin
  if NextFree <= LZW_Table_SIZE then
   begin
     i := FreeList^[NextFree];
     inc(NextFree);
     LZW_Table^[I].Prefix  := Prefix;
     LZW_Table^[I].Suffix  := Suffix;
     inc(LZW_Table^[Prefix].ChildCount);
   end {if};
end;



function ZipObj.Get_Code(CodeSize : Byte) : Integer;
const
  Mask   : array[1..8] of byte = ($01, $03, $07, $0F, $1F, $3F, $7F, $FF);
  TmpInt : integer = 0;
var
  BitsNeeded : byte;
  HowMany    : byte;
  HoldCode   : integer;
begin
  if FirstCh then
   begin                               { If first time through ...         }
     TmpInt := Get_Compressed;         { ... then prime the code buffer    }
     if TmpInt = -1 then
      begin                            { If EOF on fill attempt ...        }
        Get_Code := -1;                { ... Otherwise, return EOF      }
        exit;
      end {if};
     SaveByte := TmpInt;
     BitsLeft := 8;                    { there's now 8 bits in our buffer  }
     FirstCh  := false;
   end {if};

  BitsNeeded := CodeSize;
  HoldCode   := 0;
  while (BitsNeeded > 0) and (TmpInt <> -1) do
   begin
     if BitsNeeded >= BitsLeft then HowMany := BitsLeft { HowMany <-- Min(BitsLeft, BitsNeeded) }
                               else HowMany := BitsNeeded;

     HoldCode := HoldCode or ((SaveByte and Mask[HowMany]) shl (CodeSize - BitsNeeded));
     SaveByte := SaveByte shr HowMany;
     Dec(BitsNeeded, HowMany);
     Dec(BitsLeft, HowMany);

     if BitsLeft <= 0 then
      begin         { If no bits left in buffer ...     }
        TmpInt := Get_Compressed;         { ... then attempt to get 8 more.   }
        if TmpInt = -1 then
         begin
          if (BitsNeeded = 0) then         { If we got what we came for ... }
           Get_Code := HoldCode            { ... then return it             }
          else
           Get_Code := -1;                 { ... Otherwise, return EOF      }
          exit;
         end;
        SaveByte := TmpInt;
        BitsLeft := 8;
      end {if};
   end {while};

  if (BitsNeeded = 0) then         { If we got what we came for ... }
   Get_Code := HoldCode            { ... then return it             }
  else
   Get_Code := -1;                 { ... Otherwise, return EOF      }
end {Get_Code};


var
  _CodeSize :byte;
  _CurrCode :integer;
  _SaveCode :integer;
  _PrevCode :integer;
  _BaseChar :byte;
  gototo    :byte;


procedure ZipObj.UnShrink(count:longint);
var
  CodeSize :byte;          { Current size (in bits) of codes coming in  }
  CurrCode :integer;
  SaveCode :integer;
  PrevCode :integer;
  BaseChar :byte;
  count_2  :longint;

label pass1, pass2, pass3;

begin
  count_2:=0;

  case gototo of
    1 :begin
         CodeSize:=_CodeSize;
         CurrCode:=_CurrCode;
         SaveCode:=_SaveCode;
         PrevCode:=_PrevCode;
         BaseChar:=_BaseChar;
         goto pass1;
       end;
    2 :begin
         CodeSize:=_CodeSize;
         CurrCode:=_CurrCode;
         SaveCode:=_SaveCode;
         PrevCode:=_PrevCode;
         BaseChar:=_BaseChar;
         goto pass2;
       end;
    3 :begin
         CodeSize:=_CodeSize;
         CurrCode:=_CurrCode;
         SaveCode:=_SaveCode;
         PrevCode:=_PrevCode;
         BaseChar:=_BaseChar;
         goto pass3;
       end;
  end; {case}

  if FirstCH then
   begin
     CodeSize := MINCODESIZE;               { Start with the smallest code size }

     PrevCode := Get_Code(CodeSize);        { Get first code from file          }
     if PrevCode = -1 then                  { If EOF already, then ...          }
      exit;                                 { ... just exit without further ado }
     BaseChar := PrevCode;
     Put_Ext(BaseChar);                     { Unpack the first character        }
     inc(count_2, 1);
     if count_2>=count then
      begin
        _CodeSize:=CodeSize;
        _CurrCode:=CurrCode;
        _SaveCode:=SaveCode;
        _PrevCode:=PrevCode;
        _BaseChar:=BaseChar;
        gototo:=1;
        Exit;
      end;
   end;

pass1:
  CurrCode := Get_Code(CodeSize);        { Get next code to prime the while loop }

  while CurrCode <> -1 do
   begin          { Repeat for all compressed bytes   }
     if CurrCode = SPECIAL then
      begin     { If we've got a "special" code ... }
        CurrCode := Get_Code(CodeSize);
        case CurrCode of
          1 :                           { ... and if followed by a 1 ...    }
              inc(CodeSize);            { ... then increase code size       }
          2 :                           { ... and if followed by a 2 ...    }
              LZW_Clear_Table;          { ... clear leaf nodes in the table }
          else
           begin                        { ... if neither 1 or 2, discard    }
             writeln;
             writeln('Encountered code 256 not followed by 1 or 2!');
             writeln;
             write('Press a key to continue ...');
             readkey;
             delline;
             gotoxy(1,WhereY);
           end {else};
        end {case};
      end {then}
     else
      begin                          { Not a "special" code              }
        SaveCode := CurrCode;             { Save this code someplace safe...  }
        if CurrCode > LZW_Table_SIZE then
         ZFSResult:=310;  {Invalid code encountered!};

        if (CurrCode >= FIRSTFREE) and (LZW_Table^[CurrCode].Prefix = -1) then
         begin
           if StackIdx > LZW_Stack_SIZE then
            begin
              writeln;
              writeln('Stack Overflow (', StackIdx, ')!');
              halt;
            end {if};
           LZW_Stack^[StackIdx] := BaseChar;
           inc(StackIdx);
           CurrCode := PrevCode;
         end {if};
        while CurrCode >= FIRSTFREE do
         begin
           if StackIdx > LZW_Stack_SIZE then
            begin
              writeln;
              writeln('Stack Overflow (', StackIdx, ')!');
              halt;
            end {if};
           LZW_Stack^[StackIdx] := LZW_Table^[CurrCode].Suffix;
           inc(StackIdx);
           CurrCode := LZW_Table^[CurrCode].Prefix;
         end {while};

        BaseChar := LZW_Table^[CurrCode].Suffix; { Get last character ...   }
        Put_Ext(BaseChar);
        inc(count_2, 1);
        if count_2>=count then
         begin
           _CodeSize:=CodeSize;
           _CurrCode:=CurrCode;
           _SaveCode:=SaveCode;
           _PrevCode:=PrevCode;
           _BaseChar:=BaseChar;
           gototo:=2;
           Exit;
         end;

pass2:
        while (StackIdx > 0) do
         begin
           dec(StackIdx);
           Put_Ext(LZW_Stack^[StackIdx]);
           inc(count_2, 1);
           if count_2>=count then
            begin
              _CodeSize:=CodeSize;
              _CurrCode:=CurrCode;
              _SaveCode:=SaveCode;
              _PrevCode:=PrevCode;
              _BaseChar:=BaseChar;
              gototo:=3;
              Exit;
            end;
         end {while};                             { ... until there are none left     }

pass3:
        LZW_Add_to_Table(PrevCode, BaseChar);    { Add new entry to table      }
        PrevCode := SaveCode;
      end {if};
     CurrCode := Get_Code(CodeSize);            { Get next code from input Zippo.Stream   }
   end {while};
end;



function ZipObj.UnZip(count :longint; btg:longint):longint;
var
  C  :  integer;
  cnt1 :longint;
begin
  if btg=0 then
   begin
     Bytes_To_Go:=LocalHdr.Compressed_Size;
     FirstCh:=True;
   end
  else
   begin
     Bytes_To_Go:=btg;
     FirstCh:=false;
   end;

  ExtCount    := 0;

  if Count=0 then count:=LocalHdr.Uncompressed_Size;

  UnShrink(count);

  UnZip:=Bytes_To_Go;
end {UnZip};

(************************************************************************)
(***************** END ZIP HANDLING ROUTINES ****************************)
(************************************************************************)



(***********************************************************************
  cOMVERTS aN aRRAY oF bYTES iNTO a sTRING
*******)
function CharToPas(var C :array of Byte) : String;
var I: Integer; S: String;
begin
  I:=0;
  while (I<=High(C)) and (C[I]<>0) do Inc(I);
  S[0]:=Chr(I);
  Move(C, S[1], I);
  CharToPas:=S;
end;

(***********************************************************************
  nORMALIZE a fILENAME
*******)
function FormatFileName(s:string) : string;
var l:string; i:integer;
begin
  l:=s;
  for i:=1 to Length(s) do
   begin
     if s[i]='/' then l[i]:='\'
                 else l[i]:=UpCase(s[i]);
   end;
  FormatFileName:=l;
end;


(***********************************************************************
  aDDS a nODE aT tHE eND oF tHE lINKED lIST oF tHE .pACK's dIRECTORY
*******)
Procedure AddMEMDir(ZIPf:pZIP; data:tFileZIP);
var
  tmp, tmp2 :pMEMDir;
begin
  ZFSResult:=0;

  new(Tmp);
  if Tmp=NIL then
   begin
     ZFSResult:=257;  {Unable to create a new node}
     exit;
   end;

  if ZIPf^.FirstMemDir=NIL then
   begin
     ZIPf^.FirstMemDir:=Tmp;
     with ZIPf^.FirstMemDir^ do
      begin
        next:=NIL;
      end;
   end
  else
   begin
     tmp2:=ZIPf^.FirstMemDir;
     while tmp2^.next<>NIL do
      begin
        tmp2:=tmp2^.next;
      end;
     tmp2^.next:=tmp;
     tmp^.next:=NIL;
   end;

  with tmp^ do
   begin
     move(data, obj, sizeof(obj));
   end;
end;


(***********************************************************************
  eRASE tHE wHOLE lINKED lIST oF a .pACK dIRECTORY
*******)
Procedure KillMEMDir(ZIPf:pZIP);
var
  tmp,
  tmp2 :pMEMDir;

begin
  ZFSResult:=0;

  tmp:=ZIPf^.FirstMemDir;
  if tmp=NIL then {there aren't nothing}
   begin
     exit;
   end
  else
   begin
     While tmp<>NIL do
      begin
        tmp2:=tmp^.NEXT;
        Dispose(tmp);
        tmp:=Tmp2;
      end;
   end;
  ZIPf^.FirstMemDir:=NIL;
end;


(***********************************************************************
  aDDS a nODE tO tHE eND oF tHE pACK fILES lINKED lIST
*******)
Function AddZIP :pZIP;
var
  tmp,
  tmp2 :pZIP;

begin
  ZFSResult:=0;

  AddZIP:=NIL;
  new(Tmp);
  if Tmp=NIL then
   begin
     ZFSResult:=257;  {Unable to create a new node}
     exit;
   end;

  if FirstZIP=NIL then
   begin
     FirstZIP:=Tmp;
     FirstZIP^.next:=NIL;
   end
  else
   begin
     tmp2:=FirstZIP;
     while tmp2^.next<>NIL do
      begin
        tmp2:=tmp2^.next;
      end;
     tmp^.next:=NIL;
     tmp2^.next:=tmp;
   end;
  AddZIP:=tmp;
end;


(***********************************************************************
  eRASES a nODE fROM tHE pACKS lINKED lIST
*******)
Procedure EraseZIP(ZIPf :pZIP);
Var
  tmp :pZIP;  { Used to move through the list. }
begin
  ZFSResult:=0;

  if ZIPf = nil then
   begin
     ZFSResult:=256;  {The pointer is NIL, can not be erased}
     exit;
   end;

  if FirstZIP = nil then   { Is the list empty? }
   begin
     ZFSResult:=258;  {No nodes to be delete in this linked list}
     exit;
   end
  else
   begin
     if FirstZIP=ZIPf then
      begin
        FirstZIP:=ZIPf^.next;
        Dispose(ZIPf);
        exit;
      end;
   end;

  tmp := FirstZIP;
  While tmp^.Next <> ZIPf do
   begin
     tmp := tmp^.Next;
     if tmp=NIL then
      begin
        ZFSResult:=259;  {node not found in this linked list}
        exit;
      end;
   end;
  tmp^.Next := ZIPf^.Next;       { Point around old link. }
  Dispose(ZIPf);              { Get rid of the link..  }
end;


(***********************************************************************
  oPENS a pACK aRCHIVE aND iNITIALIZE iT
*******)
Function OpenZIP (ZIPName :string ):pZIP;
var i:integer;
    num :byte;
    ZIPf:pZIP;
    dat:tfileZip;
    w :word;
begin
{$I-}
  ZFSResult:=0;
  OpenZIP:=NIL;

  ZIPf:=AddZIP;
  if ZFSResult<>0 then Exit;

  ZIPf^.Obj.Init(ZIPName);
  if ZFSResult<>0 then
   begin
     EraseZIP(ZIPf);
     Exit;
   end;

  ZIPf^.Active:=true;
  ZIPf^.Name:=ZIPName;

  ZIPf^.NumFiles:=0;
  ZIPf^.FirstMemDir:=NIL;
  while ZIPf^.Obj.Read_Local_Hdr do
   begin
     inc(ZIPf^.NumFiles,1);

     dat.Nom:=ZIPf^.Obj.Hdr_FileName;
     dat.Kind:=ZIPf^.Obj.LocalHdr.Compress_Method;
     dat.Pos:=system.FilePos(ZIPf^.Obj.ZIPfile)+ZIPf^.obj.ZipPtr-ZIPf^.obj.ZipCount-1;
     dat.Size:=ZIPf^.Obj.LocalHdr.Compressed_Size;
     dat.USize:=ZIPf^.Obj.LocalHdr.UnCompressed_Size;
     addMemDir(ZIPf, dat);

     ZIPf^.Obj.fSkip(ZIPf^.Obj.LocalHdr.Compressed_Size);
   end;

  If ZIPf^.NumFiles<1 then
   begin
     ZIPf^.Obj.Close;
     EraseZIP(ZIPf);
     ZFSResult:=300; {Pack file without files}
     exit;
   end;

  System.Seek(ZIPf^.Obj.ZIPfile, 0);   {We go to the start of the ZIP file...}
  if IOResult<>0 then
   begin
     ZFSResult:=IOResult;
     exit;
   end;
  ZIPf^.Obj.Read_Zip_Block;            {...and we fill the InBuffer with data}


  OpenZIP:=ZIPf;
{$I+}
end;


(***********************************************************************
  cLOSES a zIP fILE
*******)
Procedure CloseZIP(ZIPf:pZIP);
begin
{$I-}
  ZFSResult:=0;

  if ZIPf=NIL then
   begin
     ZFSResult:=256;
     Exit;
   end;

  ZIPf^.Obj.Close;

  KillMEMDir(ZIPf);
  EraseZIP(ZIPf);
{$I+}
end;


(***********************************************************************
  lOOKS iF eXIST a fILE iN tHE dISK oR iN tHE oPPENED pACKS
*******)
FUNCTION FileExist( sExp :STRING ): BOOLEAN;
VAR
  s       :SearchRec;
  tmpZIP  :pZIP;
  tmpDir  :pMEMDir;
BEGIN
{$I-}
  DOS.FindFirst( sExp, Archive, s );
  if DOSError=0 then
   begin
     FileExist:=true;
     Exit;
   end;

  if FirstZIP=NIL then {No oppened PACKS}
   begin
     FileExist:=False;
     exit;
   end;

  tmpZIP:=FirstZIP;
  REPEAT

    if tmpZIP^.FirstMemDir<>NIL then {If there are any file in the PACK}
     begin
       tmpDir:=tmpZIP^.FirstMemDir;
       REPEAT
         if FormatFileName(tmpDir^.Obj.Nom)=FormatFileName(sExp) then
          begin
            FileExist:=TRUE;
            exit;
          end;

         tmpDir:=tmpDir^.Next;
       UNTIL tmpDir=NIL;
     end;

    tmpZIP:=tmpZIP^.next;
  UNTIL tmpZIP=NIL;

  FileExist := False;
{$I+}
END;


Procedure Assign(var f:ZFSfile; fname:string);
VAR
  s       :SearchRec;
  tmpZIP  :pZIP;
  tmpDir  :pMEMDir;
BEGIN
{$I-}
  ZFSResult:=0;

  DOS.FindFirst( fname, Archive, s );
  if DOSError=0 then
   begin
     system.Assign(f.handle, fname);
     system.reset(f.handle,1);
     If IOResult<>0 then
      begin
        ZFSResult:=IOResult;
        exit;
      end;

     f.InZIP:=FALSE;
     f.Openned:=TRUE;
     f.ZIP:=NIL;
     f.size:=system.filesize(f.handle);
     if IOResult<>0 then
      begin
        ZFSResult:=IOResult;
        exit;
      end;
     f.USize:=f.size;
     f.comp:=0;
     f.pos:=0;
     f.curr:=0;

     Exit;
   end;

  if FirstZIP=NIL then {No oppened PACKS}
   begin
     ZFSResult:=2; {File not found}
     exit;
   end;

  tmpZIP:=FirstZIP;
  REPEAT

    if tmpZIP^.FirstMemDir<>NIL then {If there are any file in the PACK}
     begin
       tmpDir:=tmpZIP^.FirstMemDir;
       REPEAT
         if FormatFileName(tmpDir^.Obj.Nom)=FormatFileName(fname) then
          begin
            f.InZIP:=TRUE;
            f.Openned:=TRUE;
            f.ZIP:=tmpZIP;
            f.comp:=tmpDir^.Obj.Kind;
            f.pos:=tmpDir^.Obj.Pos;
            f.size:=tmpDir^.Obj.Size;
            f.USize:=tmpDir^.Obj.USize;
            f.curr:=0;

            exit;
          end;

         tmpDir:=tmpDir^.Next;
       UNTIL tmpDir=NIL;
     end;

    tmpZIP:=tmpZIP^.next;
  UNTIL tmpZIP=NIL;

  ZFSResult:=2; {File not found}
  f.Openned:=FALSE;
{$I+}
end;


Procedure Close(var f:ZFSfile);
BEGIN
{$I-}
  ZFSResult:=0;
  if NOT f.Openned then
   begin
     ZFSResult:=303;
     exit;
   end;

  if NOT f.InZIP then
   begin
     system.Close(f.Handle);
     if IOResult<>0 then
      begin
        ZFSResult:=IOResult;
        exit;
      end;
   end;

  f.InZIP:=False;
  f.Openned:=FALSE;
  f.ZIP:=NIL;
  f.size:=0;
  f.pos:=0;
  f.curr:=0;
{$I+}
end;


procedure seek(var f:ZFSfile; N:longint);
begin
{$I-}
  ZFSResult:=0;

  if NOT f.Openned then
   begin
     ZFSResult:=303;
     exit;
   end;

  if f.InZIP then
   begin
     if f.Comp=0 then
      begin
        if (N>F.Size) or (N<0) then
         begin
           ZFSResult:=304;
           exit;
         end;

        system.seek(f.ZIP^.Obj.ZIPfile, f.POS+N);
        ZFSResult:=IOResult;
        f.ZIP^.Obj.Read_Zip_Block;
        f.Curr:=N;
      end
     else
      begin
        f.Curr:=N;
      end;
   end
  else
   begin
     system.seek(f.handle, n);
     ZFSResult:=IOResult;
     f.curr:=n;
   end;
{$I-}
end;


function BlockRead(var F :ZFSfile; var Buf; Count :longint):longint;
const
  tmpbufsize = 4096;
var
  tmpbuf:array[0..tmpbufsize-1] of byte;
  Res:longint;
  TmpDest:Pointer;
  cnt1:longint;
  tmp1:longint;
  oldbtg:longint;

begin
{$I-}
  BlockRead:=0;
  ZFSResult:=0;

  if NOT f.Openned then
   begin
     ZFSResult:=303;
     exit;
   end;

  if f.InZIP then
   begin

     if f.Comp=0 then
      begin
        if (f.curr+Count)>f.USize then
         begin
           {###ERROR### - The requested data exceeds the file limits}
           Count:=f.USize-f.Curr;
         end;

        system.Seek(f.ZIP^.Obj.ZIPfile, f.pos+f.curr);
        if IOResult<>0 then
         begin
           ZFSResult:=IOResult;
           exit;
         end;

        system.BlockRead(f.ZIP^.Obj.ZIPfile, Buf, Count, Res);
        if IOResult<>0 then
         begin
           ZFSResult:=IOResult;
           exit;
         end;
        f.curr:=f.curr+Res;
      end
     else
      begin
        if (f.curr+Count)>f.USize then
         begin
           {###ERROR### - The requested data exceeds the file limits}
           Count:=f.USize-f.Curr;
         end;

        if f.comp=1 then
         if NOT f.ZIP^.obj.LZW_Init then
          begin
            {###ERROR### - Not enough memory for init LZW tables}
          end;

        tmpdest:=f.ZIP^.obj.Dest;
        f.ZIP^.obj.Dest:=@tmpbuf;

        f.ZIP^.Obj.LocalHdr.Compress_Method:=f.comp;
        f.ZIP^.Obj.LocalHdr.Compressed_Size:=f.size;

        system.seek(f.ZIP^.Obj.ZIPfile, f.pos);
        f.ZIP^.Obj.Read_Zip_Block;

        gototo:=0;

        oldbtg:=0;
        for cnt1:=0 to (f.curr DIV TmpBufSize) do
         begin
           oldbtg:=f.ZIP^.Obj.Unzip(tmpbufsize, oldbtg);
         end;

        tmp1:=f.curr MOD TmpBufSize;

        if count<=(tmpBufSize-Tmp1) then
         begin
           Move(tmpbuf[tmp1], buf, count);
         end
        else
         begin
           Move(tmpbuf[tmp1], buf, TmpBufSize-tmp1);

           f.ZIP^.obj.Dest:=@buf;
           inc(longint(f.ZIP^.obj.Dest), (TmpBufSize-tmp1));

           oldbtg:=f.ZIP^.Obj.Unzip(count-(TmpBufSize-tmp1), oldbtg);
         end;

        f.curr:=f.curr+Count;

        if f.comp=1 then f.ZIP^.obj.LZW_Cleanup;

        f.ZIP^.obj.Dest:=tmpdest;
      end;
   end
  else
   begin
     system.BlockRead(f.handle, Buf, Count, Res);
     if IOResult<>0 then
      begin
        ZFSResult:=IOResult;
        exit;
      end;
   end;

  BlockRead:=RES;
{$I+}
end;


Function EOF(f:ZFSfile):Boolean;
begin
{$I-}
  EOF:=FALSE;

  with f do
   begin
     if Openned then
      if InZIP then EOF:=(Curr>=USize )
               else EOF:=SYSTEM.EOF(handle);
   end;

  ZFSResult:=IOResult;
{$I+}
end;



Procedure KillZIPs;
var
  tmp,
  tmp2 :pZIP;
begin
  tmp:=FirstZIP;
  if tmp=NIL then              {there is nothing to close}
   begin
   end
  else
   begin
     While tmp<>NIL do
      begin
        tmp2:=tmp^.NEXT;
        CloseZIP(tmp);
        tmp:=Tmp2;
      end;
   end;
  FirstZIP:=NIL;
end;


(***********************************************************************
  cLOSES aLL tHE oPPENED fILES aUTOMATICLY wHEN wE sTOP tHE pROGRAM
*******)
procedure ZIPExitProc; FAR;
begin
  ExitProc:=OldExitProc;        {rescue the original exit}
  KillZIPs;
end;




BEGIN
  OldExitProc:=ExitProc;
  ExitProc:=@ZIPExitProc;
END.